#include <cuda.h>
#include <cublas.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <pthread.h>
#include <assert.h>

#define SIZE	128

float *h_A, *h_B;
CUdeviceptr d_A;

pthread_t thread;

CUcontext cuContext;

void *thread_func(void *arg __attribute__((unused)))
{
	CUresult status;
	CUcontext pctx;

	printf("foo\n");

	/* get the context  */
	status = cuCtxPushCurrent(cuContext);
	assert(!status);

	/* upload the vector */
	cuMemcpyHtoD(d_A, h_A, SIZE*sizeof(float));

	/* put the context */
	status = cuCtxPopCurrent(&pctx);
	assert(!status);

	printf("(THREAD) cuCtxPopCurrent returned %p\n", pctx);


	return NULL;
}

void cuda_init(void)
{
	CUresult status;

	/* initialize CUDA and CUBLAS */
	status = cuInit(0);
	if (status)
		assert(0);

	status = cuCtxCreate(&cuContext, 0, 0);
	assert(!status);

	/* that should associate cublas with the current context */
	cublasInit();

	printf("(MAIN) created contex %p\n", cuContext);
}

int main(int argc, char **argv)
{
	CUresult status;
	cublasStatus cublas_status;

	CUcontext pctx;

	h_A = malloc(SIZE*sizeof(float));
	assert(h_A);

	h_B = malloc(SIZE*sizeof(float));
	assert(h_B);

	h_A[0] = 42.0f;

	cuda_init();

	/* malloc a buffer on the device */
	cuMemAlloc(&d_A, SIZE*sizeof(float));
	assert(d_A);

	printf("Allocate a buffer on device at %p\n", d_A);

	/* release the context */
	status = cuCtxPopCurrent(&pctx);
	assert(!status);

	printf("(MAIN) cuCtxPopCurrent returned %p\n", pctx);

	/* launch a new thread that use CUBLAS */
	pthread_create(&thread, NULL, thread_func, NULL);

	pthread_join(thread, NULL);

	/* get the context  */
	status = cuCtxPushCurrent(cuContext);
	assert(!status);

	/* use CUBLAS to multiply the vector by 2.0f */
	cublasSscal (SIZE, 2.0f, d_A, 1);
	cublas_status = cublasGetError();
	printf("cublas_status = %x\n", cublas_status);
	assert(!cublas_status);

	/* get the vector back */
	h_B[0] = -1.0f;
	cuMemcpyDtoH(h_B, d_A, SIZE*sizeof(float));

	printf("h_B[0] = %f\n", h_B[0]);

	free(h_A);
	return 0;
}
